"""
The main logic for polling, executing and reporting jobs
"""
import os
import platform
import tempfile
import shutil
from .common import ZipFileEx
from .httpclient import Session
from .consts import NAME, VERSION
from requests_toolbelt.multipart import encoder


class Runner(Session):
    """
    The Python Gitlab Runner
    """

    def __init__(self, server, token):
        """
        Start a runner
        :param server: The gitlab server
        :param token: The runner access token (not the registration token)
        """
        super(Runner, self).__init__()
        self.server = server
        self.builds = os.getcwd()
        self.token = token
        self.api_prefix = "{}/api/v4".format(self.server)
        self.executor = "shell"
        self.shell = "bash"
        if platform.system() == "Windows":
            self.shell = "cmd"

        self.image = False
        self.services = False

    def features(self):
        """
        Get the features this runner is enabled with
        :return:
        """
        return {
            "variables": True,
            "artifacts": True,
            "artifacts_exclude": True,
            "cache": False,
            "shared": True,
            "upload_multiple_artifacts": True,
            "session": True,
            "terminal": True,
            "image": self.image,  # set True for docker later
            "services": self.services,
        }

    def api(self, name):
        """
        Return the named API endpoint
        :param name:
        :return:
        """
        return "{}/{}".format(self.api_prefix, name)

    def register(self, desc, regtoken, tags=["gitlab-python-runner_added"]):
        """
        Register this runner with the server
        :param desc: The description for this new runner
        :param regtoken: The runner registration token
        :param tags: set these runner tags
        :return:
        """
        resp = self.post(self.api("runners"), data={
            "token": regtoken,
            "description": desc,
            "run_untagged": False,
            "tag_list": tags
        })

        resp.raise_for_status()

        result = self.json(resp)

        self.token = result.get("token", None)

        assert self.token, "Bug? Failed to register, no token found!"

    def unregister(self, token=None):
        """
        Delete a registered runner
        :param token:
        :return:
        """
        if token:
            self.token = token

        resp = self.delete(self.api("runners"), data={
            "token": self.token
        })

        resp.raise_for_status()

        self.token = None

    def get_info(self):
        """
        Return the info dict
        :return:
        """
        opsys = platform.system().lower()

        cpu = platform.machine()

        if cpu == "x86_64":
            cpu = "amd64"

        return {
            "name": NAME,
            "version": VERSION,
            "revision": VERSION,
            "platform": opsys,
            "architecture": cpu,
            "executor": self.executor,
            "shell": self.shell,
            "features": self.features(),
        }

    def poll(self):
        """
        Ask gitlab if there are any jobs we can run
        :return:
        """
        resp = self.post(self.api("jobs/request"), json={
            "info": self.get_info(),
            "token": self.token,
        })
        resp.raise_for_status()
        if resp.status_code == 201:
            # we have been given a job!
            return self.json(resp)

        return None

    def trace(self, trace, text, offset=0):
        """
        Upload some trace
        :param trace:
        :param text:
        :param offset:
        :return:
        """
        job = trace.job
        resp = self.patch(self.api("jobs/{}/trace".format(job["id"])),
                          headers={
                              "Content-Type": "text/plain",
                              "Content-Length": str(len(text)),
                              "Content-Range": "{}-{}".format(offset, offset + len(text)),
                              "Job-Token": job["token"]},
                          data=text)
        resp.raise_for_status()

        return offset + len(text)

    def success(self, job):
        """
        Report success
        :param job:
        :return:
        """
        resp = self.put(self.api("jobs/{}".format(job["id"])), json={
            "info": self.get_info(),
            "token": job["token"],
            "state": "success",
        })

        resp.raise_for_status()

    def failed(self, job):
        """
        Report failure
        :param job:
        :return:
        """
        resp = self.put(self.api("jobs/{}".format(job["id"])), json={
            "info": self.get_info(),
            "token": job["token"],
            "state": "failed",
        })
        resp.raise_for_status()

    def get_dependencies(self, trace, job, build_dir):
        """
        Get the dependencies for the given job
        :param trace: trace
        :param job: gitlab job object
        :param build_dir: Place to unpack artifacts into
        :return:
        """
        for other in job["dependencies"]:
            if "artifacts_file" not in other:  # other job did not save any artifacts
                continue
            trace.writeline("Fetching artifacts from {}..".format(other["name"]))
            url = self.api("jobs/{}/artifacts".format(other["id"]))
            resp = self.get(url,
                            stream=True,
                            headers={
                                "Job-Token": other["token"],
                            })
            resp.raise_for_status()
            try:
                tmpdir = tempfile.mkdtemp()
                path = os.path.join(tmpdir, other["artifacts_file"]["filename"])
                with open(path, 'wb') as f:
                    resp.raw.decode_content = True
                    shutil.copyfileobj(resp.raw, f)
                self.unpack_dependencies(trace, path, build_dir)
            finally:
                shutil.rmtree(tmpdir)

    def unpack_dependencies(self, trace, archive_file, build_dir):
        """
        Extract a downloaded artifact into the build tree
        :param trace:
        :param archive_file:
        :param build_dir:
        :return:
        """
        if self.features().get("artifacts", False):
            try:
                tmpdir = tempfile.mkdtemp()
                with ZipFileEx(archive_file, mode="r") as zf:
                    trace.writeline("Unpacking artifacts into {}..".format(build_dir))
                    zf.extractall(path=build_dir)
            finally:
                shutil.rmtree(tmpdir)

    def upload(self, trace, job, archive_file):
        """
        Upload artifacts from a job
        :param trace:
        :param job:
        :param archive_file:
        :return:
        """
        url = self.api("jobs/{}/artifacts".format(job["id"]))
        trace.writeline("Uploading artifacts to gitlab..")
        with open(archive_file, "rb") as archive:
            # solution to large uploads
            # lifted from https://stackoverflow.com/questions/35779879/python-requests-upload-large-file-with-additional-data
            form = encoder.MultipartEncoder({
                "file": (os.path.basename(archive_file), archive, "application/octet-stream"),
                "artifact_format": "zip",
                "artifact_type": "archive"
            })
            headers = {
                       "Content-Type": form.content_type,
                       "Job-Token": job["token"]
                       }
            resp = self.post(url,
                             headers=headers,
                             data=form)
        resp.raise_for_status()
        trace.writeline("Upload complete")


class DockerRunner(Runner):
    """
    Docker Enabled mode
    """
    def __init__(self, server, token):
        super(DockerRunner, self).__init__(server, token)
        self.image = True
        self.services = True



"""
File-like trace bufffer that sends to gitlab
"""
import os
import time


class TraceProxy(object):
    """
    Handle sending trace from the runner to gitlab
    """

    def __init__(self, runner, job):
        self.runner = runner
        self.job = job
        self.offset = 0
        self.emulator_job = None

        self.buffer = b""
        self.last_write = 0
        self.write_interval = 2

    def _should_write(self):
        """
        Return True if we should write data to the job trace
        :param data:
        :return:
        """
        if len(self.buffer) > 1024:
            return True
        if time.time() - self.last_write > self.write_interval:
            return True
        return False

    def write(self, data, flush=False):
        """
        Log the given data to the server
        :param data:
        :param flush: write any buffered content
        :return:
        """
        if hasattr(data, "encode"):
            data = data.encode()
        self.buffer += data

        if self._should_write() or flush:
            self.flush()

    def writeline(self, text):
        """
        Write a message on it's own line
        :param text:
        :return:
        """
        text = str(text)
        if not text.endswith(os.linesep):
            text += os.linesep
        self.write(text)

    def flush(self):
        """
        Flush the buffer
        :return:
        """
        self.offset = self.runner.trace(self, self.buffer, offset=self.offset)
        self.buffer = b""
        self.last_write = time.time()

    def close(self):
        """
        Close the buffer
        :return:
        """
        pass



#!/usr/bin/python
"""
A Pure Python gitlab runner
"""

import argparse

import os
import sys

import time
import yaml
from requests.exceptions import ConnectionError, HTTPError
from . import consts, runner, executor, common, service
from gitlabemu import logmsg


parser = argparse.ArgumentParser(prog="{} -m GitlabPyRunner".format(sys.executable))
parser.add_argument("--register",
                    type=str,
                    help="Register a new runner with the server",
                    )
parser.add_argument("--regtoken",
                    type=str,
                    default=None,
                    help="Registration token for --register")
parser.add_argument("--unregister",
                    default=False, action="store_true",
                    help="De-register the runner")

parser.add_argument("--type", type=str,
                    default="shell",
                    help="Set the runner executor eg(shell, docker)")

parser.add_argument("--shell", type=str,
                    help="Set the executor shell")

parser.add_argument("--desc", type=str,
                    help="Set the runner description")

parser.add_argument("--tag", type=str, action="append",
                    help="Add a tag when registering a runner",
                    )
parser.add_argument("--start", type=str,
                    help="Start the runner defined in the config file"
                    )

parser.add_argument("--trace-http", default=False, action="store_true", dest="httpdebug",
                    help="Log HTTP requests and responses for debugging purposes only"
                    )

parser.add_argument("--once", action="store_true", default=False,
                    help="Run only one job and then exit"
                    )

installer = service.get_installer()

if installer is not None:
    parser.add_argument("--install", type=str, default=None, metavar="CONFIG", dest="install",
                        help="Launch gitlab-runner.py on boot using the given CONFIG file")

    parser.add_argument("--user", type=str, default=None,
                        help="User to use for --install")

    parser.add_argument("--uninstall", action="store_true", default=False, dest="uninstall",
                        help="Remove boot-time startup scripts (undo --install)")


def run():
    opts = parser.parse_args()

    if installer:
        if opts.uninstall:
            installer.uninstall()
            return
        if opts.install:
            installer.install(opts.install, opts.user)
            return

    if opts.register:
        if not opts.desc:
            opts.desc = consts.NAME + " on " + common.gethostname()

        if not opts.tag:
            opts.tag = ["new-python-runner-" + common.gethostname()]

        if not opts.regtoken:
            raise RuntimeError("missing required --regtoken")

        if not opts.shell:
            opts.shell = os.getenv("COMSPEC", "/bin/sh")

        instance = runner.Runner(opts.register, None)
        if opts.shell:
            os.environ["SHELL"] = opts.shell  # TODO pass this into GLE better
        instance.register(opts.desc, opts.regtoken, opts.tag)

        if not instance.token:
            raise RuntimeError("runner register failed")

        # save the runner config
        tosave = {
            "server": opts.register,
            "token": instance.token,
            "executor": opts.type,
            "shell": opts.shell,
            "dir": os.getcwd()
        }

        common.save_config("gitlab-runner.yml", tosave)

        logmsg.info("Registration complete. Config saved at '{}'".format(os.path.join(os.getcwd(),
                                                                                      "gitlab-runner.yml")))
        sys.exit(0)

    if opts.unregister:
        config = common.parse_config("gitlab-runner.yml")
        logmsg.info("De-registering..")
        instance = runner.Runner(config["server"], config["token"])
        instance.unregister()
        logmsg.info("De-register complete")
        sys.exit(0)

    if opts.start:
        if opts.httpdebug:
            common.set_trace_http_requests(True)

        config = common.parse_config(opts.start)
        os.chdir(config["dir"])
        extype = config["executor"]

        if opts.once:
            logmsg.info("Will exit after one job")

        exitstatus = 1

        docker = False

        # loop forever to cope with some common errors that can come up
        while True:
            if extype == "shell":
                instance = runner.Runner(config["server"], config["token"])
            elif extype == "docker":
                docker = True
                instance = runner.DockerRunner(config["server"], config["token"])
            else:
                raise RuntimeError("unsupported executor type '{}'".format(extype))

            try:
                instance.shell = config["shell"]
                while True:
                    exitstatus = 1
                    result = None
                    interval = 40
                    logmsg.info("Polling for jobs.. ({} sec)".format(interval))
                    job = instance.poll()
                    if not job:
                        time.sleep(interval)
                    else:
                        logmsg.info("Got new job.")
                        try:
                            result = executor.run(instance, job, docker)
                            if result:
                                exitstatus = 0
                        except Exception as rerr:
                            logmsg.warning("Executor had exception {}".format(rerr))

                        finally:
                            logmsg.info("Job {} has ended success={}".format(job["id"], result))
                            if result:
                                instance.success(job)
                            else:
                                if result is None:
                                    logmsg.warning("Job {} failed early setup, see job logs".format(job["id"]))
                                instance.failed(job)
                            if opts.once:
                                break

            except (ConnectionError, HTTPError) as err:
                aborted = False
                if not result:
                    if job:
                        if err.response.status_code == 403:
                            # job was aborted by the server
                            aborted = True
                if aborted:
                    logmsg.info("Job {} was cancelled".format(job["id"]))
                else:
                    logmsg.warning("comms exception {}".format(err))
                    logmsg.warning("sleeping..")
                    time.sleep(10)
            finally:
                if opts.once:
                    break

        sys.exit(exitstatus)
